/* $Id$ */

/*========================================================================
 *  Copyright (c) Michael J. Hammel 1996.
 *========================================================================
 *              FILE NAME: coils.c
 *            DESCRIPTION: Creates a coil (spring) out of torii for POV-Ray
 *                  NOTES: set tabstops to 3 for better readability
 *
 * SPECIAL CONSIDERATIONS:
 * This program is freely redistributable as long as this copyright block
 * is left intact and that any changes you make are clearly marked as
 * yours.  Changes should be noted in the Modification History below.
 *
 *========================================================================
 *
 * MODIFICATION HISTORY:
 * $Log$
 *
 *========================================================================*/

/* === System Headers === */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <pwd.h>
#include <time.h>

/* === Project Headers === */
#include "coils.h"

/* === external routines === */

/* === Public routine prototypes === */
void DoPovray(double rotation_angle);

/* === Private routine prototypes === */

/* === Global Variables === */
char		*Filename;
int		Format_Type;
char		*ObjectName;
int		Append;

int		Coil_Rotations;
double	Coil_Tightness;
double	Coil_Width;
double	Coil_Rotation_Width;
int		Coil_Sturm;

struct passwd	*user;
struct tm		*tm_tp;


/*========================================================================
 *	Name:			ParseCmdLine
 *	Prototype:	ParseCmdLine( int argc, char **argv )
 *					
 *	Description:
 * parser for command line arguments
 *
 *========================================================================*/
int 
ParseCmdLine(
	int argc,
	char **argv
)
{
	int	c;

	opterr=0;
	while ( (c=getopt(argc, argv, "?asf:N:o:r:t:w:W:")) != -1 )
	{
		switch (c)
		{
			/* options for any of my utilities */
			case 'a':	Append = True;
							break;
			case 'f':	Filename = (char *) malloc (strlen(optarg) + 1);
							strcpy(Filename,optarg);
							break;
			case 'N':	ObjectName = (char *) malloc (strlen(optarg) + 1);
							strcpy(ObjectName,optarg);
							break;
			case 'o':	Format_Type = atoi(optarg);
							break;

			/* Coil specific options */
			case 'r':	Coil_Rotations = atoi(optarg);
							break;
			case 't':	Coil_Tightness = atof(optarg);
							break;
			case 'w':	Coil_Width = atof(optarg);
							break;
			case 'W':	Coil_Rotation_Width = atof(optarg);
							break;
			case 's':	Coil_Sturm = True;
							break;

			case '?':	
			default:		fprintf(stderr, "%s %s", argv[0], USAGE);
							exit(0);
							break;
		}
	}
	return(optind);
}


int
main(
	int	argc,
	char	**argv
)
{
	double	rotation_angle;

	int		uid;
	time_t	time_tp;


	/* Initializiation */
	Filename = NULL;
	ObjectName = NULL;
	Format_Type = 1;
	Append = False;

	Coil_Rotations = 4;
	Coil_Tightness = 0.5;
	Coil_Width = 0.1;
	Coil_Rotation_Width = 1.0;
	Coil_Sturm = False;

	/* get the users name */
	uid = (int)getuid();
	user = getpwuid(uid);

	/* get the current date */
	time(&time_tp);
	tm_tp = localtime(&time_tp);
	
	/*
	 * Check for user specified parameters
	 */
	ParseCmdLine( argc, argv );

	/*
	 * Error check the input values
	 */
	if ( Coil_Tightness < (2 * Coil_Width) )
	{
		fprintf(stderr, 
			"Coil warning: The tightness should be greater than 2*coil-width or "
			"the coils may overlap.  This is not enforced, so we'll "
			"continue.\n");
	}

	if ( Coil_Rotation_Width <= Coil_Width )
	{
		fprintf(stderr, 
			"Coil error: The coil rotation width should be greater "
			"than the coil width or POV-Ray will report an error when you "
			"try to render the coil.  We can't ignore this. "
			"Please try again.\n");
		exit (1);
	}

	/* Let the user know whats about to happen */
	printf("Coil Rotations           : %d\n", Coil_Rotations);
	printf("Coil Tightness           : %f\n", Coil_Tightness);
	printf("Coil Width               : %f\n", Coil_Width);
	printf("Coil Rotation Width      : %f\n", Coil_Rotation_Width);
	printf("Coil Sturm               : ");
	if ( Coil_Sturm )
		printf( "On\n" );
	else
		printf( "Off\n" );
	printf("Format Type              : ");
	switch(Format_Type)
	{
		case POVRAY:	printf("POVRay\n"); break;
		default:			printf("Uknown type\n"); break;
	}
	if ( Filename == NULL )
		printf("Filename                 : coils.inc ");
	else
		printf("Filename                 : %s ", Filename);
	if ( Append )
		printf("(appending to file)\n");
	else
		printf("\n");
	if ( ObjectName == NULL )
		printf("ObjectName               : coil\n");
	else
		printf("ObjectName               : %s\n", ObjectName);

	/*
	 * Now do some work...
	 */

	/*
	 * Computer the angle that the half torii will be rotated.
	 * The angle is computed in degrees.
	 */
	rotation_angle = (180/PI) * 
			asin( (Coil_Tightness / 2.0) / (Coil_Rotation_Width * 2.0) );

	
	/*
	 * Do format specific output
	 */
	switch (Format_Type)
	{
		case POVRAY:
			DoPovray(rotation_angle);
			break;
		default:
			printf("Unknown output format type: %d\n", Format_Type);
			exit (1);
	}

	return(0);
}



/*========================================================================
 *	Name:			DoPovray
 *	Prototype:	DoPovray( double rotation_angle )
 *					
 *	Description:
 * output the coil in POV-Ray format
 *
 *========================================================================*/
void
DoPovray(
	double	rotation_angle
)
{
	FILE		*fd;
	char		*writetype;
	int		i;
	double	offset;
	char		obj_name[1024];
	
	/*
	 * We need to get the object name, so we don't do this over and over
	 * again.
	 */
	if ( ObjectName == NULL )
		sprintf(obj_name,"coil");
	else
		sprintf(obj_name,ObjectName);

	/*
	 * Open the output file
	 */
	if ( Append )
		writetype = "a";
	else
		writetype = "w";
		
	if ( Filename == NULL )
	{
		fd = fopen("coils.inc", writetype);
		if ( fd == NULL )
			printf("Failed to open output file coils.inc.\n"
				 "errno = %d\n", errno);
	}
	else
	{
		fd = fopen(Filename, writetype);
		if ( fd == NULL )
			printf("Failed to open output file %s.\n"
				 "errno = %d\n", Filename, errno);
	}

	/*
	 * Output file header info
	 */
	fprintf( fd, "%s\n", HEADER_1 );
	fprintf( fd, "%s\n", HEADER_SPACE );
	if ( !Append )
	{
		if ( Filename == NULL )
			fprintf( fd, "%s Filename                 : coils.inc\n", HEADER_SPACE);
		else
			fprintf( fd, "%s Filename                 : %s\n", HEADER_SPACE, Filename);
	}
	fprintf( fd, "%s Author                   : %s\n", HEADER_SPACE, user->pw_name );
	fprintf( fd, "%s Date                     : %02d/%02d/%02d\n", 
			HEADER_SPACE, tm_tp->tm_mon, tm_tp->tm_mday, tm_tp->tm_year );
	fprintf( fd, "%s\n", HEADER_SPACE );
	
	fprintf( fd, "%s Coil Rotations           : %d\n", HEADER_SPACE, Coil_Rotations);
	fprintf( fd, "%s Coil Tightness           : %f\n", HEADER_SPACE, Coil_Tightness);
	fprintf( fd, "%s Coil Width               : %f\n", HEADER_SPACE, Coil_Width);
	fprintf( fd, "%s Coil Rotation Width      : %f\n", HEADER_SPACE, Coil_Rotation_Width);
	fprintf( fd, "%s Coil Sturm               : ", HEADER_SPACE );
	if ( Coil_Sturm )
		fprintf( fd, "On\n" );
	else
		fprintf( fd, "Off\n" );
	fprintf( fd, "%s Start Point Coordinates  : <-%f, %f, 0>\n", 
			HEADER_SPACE, Coil_Rotations * Coil_Tightness / 2, Coil_Rotation_Width);
	fprintf( fd, "%s End Point Coordinates    : < %f, %f, 0>\n", 
			HEADER_SPACE, Coil_Rotations * Coil_Tightness / 2, Coil_Rotation_Width);


	fprintf( fd, "%s Format Type              : POVRay\n", HEADER_SPACE );
	fprintf( fd, "%s ObjectName               : %s\n", HEADER_SPACE, obj_name);

	fprintf( fd, "%s\n", HEADER_SPACE );
	if ( !Append )
	{
		fprintf( fd, "%s File automatically generated by Coils program.\n", HEADER_SPACE );
		fprintf( fd, "%s Coils program Copyright 1996 by Michael J. Hammel\n", HEADER_SPACE );
		fprintf( fd, "%s\n", HEADER_SPACE );
	}
	fprintf( fd, "%s\n", HEADER_2 );

	/*
	 * Create two objects.  Each is half of a full rotation and each is rotated
	 * by the rotation_angle in opposite directions around the z axis.
	 * Each full rotation of a coil is made up by connecting the endpoints
	 * of these two objects and translating to the appropriate position
	 * in the full coil.
	 */

	/* Left hand side of coil (+z side of torus) */
	fprintf(fd, "#declare %s_lhs ", obj_name);
	fprintf(fd, "= difference {\n");
	fprintf(fd, "%s torus {%f, %f rotate <0,0,90> ", 
			HEADER_TAB, Coil_Rotation_Width, Coil_Width );
	if ( Coil_Sturm )
		fprintf(fd, "sturm \n");
	fprintf(fd, "}\n");
	fprintf(fd, "%s plane { -z, 0 }\n", HEADER_TAB );
	fprintf(fd, "%s translate <0,-%f,0>\n", HEADER_TAB, Coil_Rotation_Width );
	fprintf(fd, "%s rotate <0,0,%f>\n", HEADER_TAB, rotation_angle );
	fprintf(fd, "}\n");

	/* Right hand side of coil (-z side of torus) */
	fprintf(fd, "#declare %s_rhs ", obj_name);
	fprintf(fd, "= difference {\n");
	fprintf(fd, "%s torus {%f, %f rotate <0,0,90>}\n", 
			HEADER_TAB, Coil_Rotation_Width, Coil_Width );
	fprintf(fd, "%s plane { z, 0 }\n", HEADER_TAB );
	fprintf(fd, "%s translate <0,-%f,0>\n", HEADER_TAB, Coil_Rotation_Width );
	fprintf(fd, "%s rotate <0,0,-%f>\n", HEADER_TAB, rotation_angle );
	fprintf(fd, "}\n");

	/*
	 * Now put these two objects together to create one full rotation of
	 * the coil.  This object is anchored with the lhs endpoint on the origin.
	 */
	fprintf(fd, "#declare %s_unit ", obj_name);
	fprintf(fd, "= union {\n");
	fprintf(fd, "%s object {%s_lhs}\n", HEADER_TAB, obj_name);
	fprintf(fd, "%s object {%s_rhs translate <%f,0,0>}\n", 
			HEADER_TAB, obj_name, Coil_Tightness);
	fprintf(fd, "}\n");


	/*
	 * The declare statement for the actual coil object
	 */
	fprintf(fd, "#declare %s ", obj_name);
	fprintf(fd, "= union {\n");

	/* 
	 * Create a unit coil for each rotation, and position it in relation
	 * to the rest of the unit coils.
	 */
	for ( i=0; i < Coil_Rotations; i++ )
	{
		if ( i == 0 )
			fprintf(fd, "%s object {%s_unit}\n", HEADER_TAB, obj_name);
		else
		{
			/* compute the offset from the origin for the current unit coil */
			offset = i * Coil_Tightness;
			fprintf(fd, "%s object {%s_unit translate <%f,0,0> }\n", 
				HEADER_TAB, obj_name, offset);
		}
	}

	/*
	 * Now center the whole thing about the origin
	 */
	offset = Coil_Rotations * Coil_Tightness / 2.0;
	fprintf(fd, "%s translate <-%f,%f,0>\n", 
			HEADER_TAB, offset, Coil_Rotation_Width );

	/*
	 * Put a bounding box around it
	 */
	fprintf(fd, "%s bounded_by {\n", HEADER_TAB);
	fprintf(fd, "%s%s box {<-%f,-%f,-%f>, <%f,%f,%f> }\n", 
			HEADER_TAB, HEADER_TAB,
			offset, Coil_Rotation_Width, Coil_Rotation_Width,
			offset, Coil_Rotation_Width, Coil_Rotation_Width );
	fprintf(fd, "%s }\n", HEADER_TAB);
	

	/*
	 * Close the object
	 */
	fprintf(fd, "}\n");

	/*
	 * Close the output file
	 */
	fclose(fd);
	return;
}
