/* $Id$ */

/*========================================================================
 *  Copyright (c) Michael J. Hammel 1996.
 *========================================================================
 *              FILE NAME: gears.c
 *            DESCRIPTION: Creates gears out of triangle meshes for POVRay
 *                  NOTES: set tabstops to 3 for better readability
 *
 * SPECIAL CONSIDERATIONS:
 * This program is freely redistributable as long as this copyright block
 * is left intact and that any changes you make are clearly marked as
 * yours.  Changes should be noted in the Modification History below.
 *
 *========================================================================
 *
 * MODIFICATION HISTORY:
 * $Log$
 *
 *========================================================================*/

/* === System Headers === */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <pwd.h>
#include <time.h>

/* === Project Headers === */
#include "gears.h"

/* === external routines === */

/* === Public routine prototypes === */
void DoPovray(Points *A0, Points *A1);

/* === Private routine prototypes === */

/* === Global Variables === */
double	Gear_Center_Cutout_Radius;
int		Gear_Number_Of_Teeth;
double	Gear_Teeth_Height;
double	Gear_Width;
double	Gear_Radius;
double	Gear_Hub_Radius;
int		Gear_Spoke_Number;
char		*Filename;
int		Format_Type;
char		*ObjectName;
int		Append;

struct passwd	*user;
struct tm		*tm_tp;


/*========================================================================
 *	Name:			ParseCmdLine
 *	Prototype:	ParseCmdLine( int argc, char **argv )
 *					
 *	Description:
 * parser for command line arguments
 *
 *========================================================================*/
int 
ParseCmdLine(
	int argc,
	char **argv
)
{
	int	c;

	opterr=0;
	while ( (c=getopt(argc, argv, "?ac:h:H:n:r:s:w:f:o:N:")) != -1 )
	{
		switch (c)
		{
			case 'a':	Append = True;
							break;
			case 'c':	Gear_Center_Cutout_Radius = atof(optarg);
							break;
			case 'f':	Filename = (char *) malloc (strlen(optarg) + 1);
							strcpy(Filename,optarg);
							break;
			case 'h':	Gear_Teeth_Height = atof(optarg);
							break;
			case 'H':	Gear_Hub_Radius = atof(optarg);
							break;
			case 'n':	Gear_Number_Of_Teeth = atoi(optarg);
							break;
			case 'N':	ObjectName = (char *) malloc (strlen(optarg) + 1);
							strcpy(ObjectName,optarg);
							break;
			case 'o':	Format_Type = atoi(optarg);
							break;
			case 'r':	Gear_Radius = atof(optarg);
							break;
			case 's':	Gear_Spoke_Number = atoi(optarg);
							break;
			case 'w':	Gear_Width = atof(optarg);
							break;

			case '?':	
			default:		fprintf(stderr, "%s %s", argv[0], USAGE);
							exit(0);
							break;
		}
	}
	return(optind);
}


int
main(
	int	argc,
	char	**argv
)
{
	int		pt;
	double	internal_angle, pt_angle;
	Points	*A0, *A0_link, *A0_last_link;
	Points	*A1, *A1_link, *A1_last_link;

	int		uid;
	time_t	time_tp;


	/* Initializiation */
	Gear_Width = 0.1;
	Gear_Teeth_Height = 0.1;
	Gear_Number_Of_Teeth = 9;
	Gear_Center_Cutout_Radius = 0.0;
	Gear_Radius = 1.0;
	Gear_Spoke_Number = 2;
	Gear_Hub_Radius = 0.0;
	Filename = NULL;
	ObjectName = NULL;
	Format_Type = 1;
	Append = False;

	/* get the users name */
	uid = (int)getuid();
	user = getpwuid(uid);

	/* get the current date */
	time(&time_tp);
	tm_tp = localtime(&time_tp);
	
	/*
	 * Check for user specified parameters
	 */
	ParseCmdLine( argc, argv );

	/*
	 * Error check the input values
	 */
	if ( Gear_Center_Cutout_Radius != 0.0 )
	{
		if ( Gear_Hub_Radius == 0.0 )
		{
			printf("You specified a Center Cutout, but did not provide a Hub Radius\n"
					"You're going to have a gear with a big whole in the middle.\n");
		}
	}

	/* Let the user know whats about to happen */
	printf("Gear Radius              : %f\n", Gear_Radius);
	printf("Gear Width               : %f\n", Gear_Width);
	printf("Gear Teeth Height        : %f\n", Gear_Teeth_Height);
	printf("Gear Number Of Teeth     : %d\n", Gear_Number_Of_Teeth);
	printf("Gear Center Cutout Radius: %f\n", Gear_Center_Cutout_Radius);
	printf("Gear Hub Radius          : %f\n", Gear_Hub_Radius);
	printf("Gear Spoke Count         : %d\n", Gear_Spoke_Number);
	printf("Format Type              : ");
	switch(Format_Type)
	{
		case POVRAY:	printf("POVRay\n"); break;
		default:			printf("Uknown type\n"); break;
	}
	if ( Filename == NULL )
		printf("Filename                 : gears.inc ");
	else
		printf("Filename                 : %s ", Filename);
	if ( Append )
		printf("(appending to file)\n");
	else
		printf("\n");
	if ( ObjectName == NULL )
		printf("ObjectName               : gear\n");
	else
		printf("ObjectName               : %s\n", ObjectName);

	/*
	 * Now do some work...
	 */

	/*
	 * Fill in an array of points in the first quadrant that specifies
	 * the bottom two points of each tooth and another array which
	 * specifies the apex of each tooth.  We only compute points in 2D
	 * since the Z portion can be determined later from the width of the
	 * gear.
	 */

	/*
	 * Just compute the first quadrant - the others can be done via
	 * symmetry.
	 */
	A0 = A0_link = A0_last_link = NULL;
	A1 = A1_link = A1_last_link = NULL;
	pt_angle = 90.0 / (double)Gear_Number_Of_Teeth;

	printf("Processing...\n");


	/* The bottom two points of each tooth */

	for ( pt=0; pt<Gear_Number_Of_Teeth+1; pt++ )
	{
		internal_angle = 90.0 - (pt * pt_angle );

		if ( A0 == NULL )
		{
			A0 = (Points *) malloc ( sizeof (Points) );
			A0_last_link = A0;
			A0_link = A0;
		}
		else
		{
			A0_link = (Points *) malloc ( sizeof (Points) );
			A0_last_link->next = A0_link;
			A0_last_link = A0_link;
		}

		A0_link->x = Gear_Radius * cos(internal_angle*(PI/180));
		A0_link->y = Gear_Radius * sin(internal_angle*(PI/180));
		A0_link->next = NULL;
	}


	/* The tops of each tooth */

	for ( pt=1; pt<Gear_Number_Of_Teeth+1; pt++ )
	{
		internal_angle = 90.0 - ((double)pt * pt_angle) + (pt_angle/2.0) ;

		if ( A1 == NULL )
		{
			A1 = (Points *) malloc ( sizeof (Points) );
			A1_last_link = A1;
			A1_link = A1;
		}
		else
		{
			A1_link = (Points *) malloc ( sizeof (Points) );
			A1_last_link->next = A1_link;
			A1_last_link = A1_link;
		}

		A1_link->x = (Gear_Radius + Gear_Teeth_Height) * cos( internal_angle*(PI/180) );
		A1_link->y = (Gear_Radius + Gear_Teeth_Height) * sin( internal_angle*(PI/180) );
		A1_link->next = NULL;

	}

	/*
	 * Do format specific output
	 */
	switch (Format_Type)
	{
		case POVRAY:
			DoPovray(A0, A1);
			break;
		default:
			printf("Unknown output format type: %d\n", Format_Type);
			exit (1);
	}

	return(0);
}


/*
 * Output in POVRay format
 */
void
DoPovray(
	Points	*A0,
	Points	*A1
)
{
	FILE		*fd;
	double	zoffset;
	Points	*A0_link, *A1_link;
	int		pt;
	char		*writetype;
	int		i;
	double	rotation_amount;

	printf("Printing");

	/*
	 * Open the output file
	 */
	if ( Append )
		writetype = "a";
	else
		writetype = "w";
		
	if ( Filename == NULL )
	{
		fd = fopen("gears.inc", writetype);
		if ( fd == NULL )
			printf("Failed to open output file gears.inc.\n"
				 "errno = %d\n", errno);
	}
	else
	{
		fd = fopen(Filename, writetype);
		if ( fd == NULL )
			printf("Failed to open output file %s.\n"
				 "errno = %d\n", Filename, errno);
	}

	/*
	 * Output file header info
	 */
	fprintf( fd, "%s\n", HEADER_1 );
	fprintf( fd, "%s\n", HEADER_SPACE );
	if ( !Append )
	{
		if ( Filename == NULL )
			fprintf( fd, "%s Filename                 : gears.inc\n", HEADER_SPACE);
		else
			fprintf( fd, "%s Filename                 : %s\n", HEADER_SPACE, Filename);
	}
	fprintf( fd, "%s Author                   : %s\n", HEADER_SPACE, user->pw_name );
	fprintf( fd, "%s Date                     : %02d/%02d/%02d\n", 
			HEADER_SPACE, tm_tp->tm_mon, tm_tp->tm_mday, tm_tp->tm_year );
	fprintf( fd, "%s\n", HEADER_SPACE );
	
	fprintf( fd, "%s Gear Radius              : %f\n", HEADER_SPACE, Gear_Radius);
	fprintf( fd, "%s Gear Width               : %f\n", HEADER_SPACE, Gear_Width);
	fprintf( fd, "%s Gear Teeth Height        : %f\n", HEADER_SPACE, Gear_Teeth_Height);
	fprintf( fd, "%s Gear Number Of Teeth     : %d\n", HEADER_SPACE, Gear_Number_Of_Teeth);
	fprintf( fd, "%s Gear Center Cutout Radius: %f\n", HEADER_SPACE, Gear_Center_Cutout_Radius);
	fprintf( fd, "%s Gear Hub Radius          : %f\n", HEADER_SPACE, Gear_Hub_Radius);
	fprintf( fd, "%s Gear Spoke Count         : %d\n", HEADER_SPACE, Gear_Spoke_Number);
	fprintf( fd, "%s Format Type              : POVRay\n", HEADER_SPACE );
	if ( ObjectName == NULL )
		fprintf( fd,"%s ObjectName               : gear\n", HEADER_SPACE );
	else
		fprintf( fd,"%s ObjectName               : %s\n", HEADER_SPACE, ObjectName);

	fprintf( fd, "%s\n", HEADER_SPACE );
	if ( !Append )
	{
		fprintf( fd, "%s File automatically generated by Gears program.\n", HEADER_SPACE );
		fprintf( fd, "%s Gears program Copyright 1996 by Michael J. Hammel\n", HEADER_SPACE );
		fprintf( fd, "%s\n", HEADER_SPACE );
	}
	fprintf( fd, "%s\n", HEADER_2 );

	/*
	 * The declare statement
	 */
	fprintf(fd, "#declare ");
	if ( ObjectName == NULL )
		fprintf(fd, "gear ");
	else
		fprintf(fd, "%s ", ObjectName);
	fprintf(fd, "= union {\n");

	/*
	 * The zoffset for all points
	 */
	zoffset = Gear_Width/2.0;

	/*
	 * The center piece
	 */
	if ( Gear_Center_Cutout_Radius != 0.0 )
	{
		fprintf(fd, "\t// Center Cutout\n");
		fprintf(fd, "\tdifference {\n");
		fprintf(fd, "\t\tcylinder { <0,0,%f>, <0,0,-%f>, %f}\n", 
			zoffset, zoffset, Gear_Radius );

		/* The center cutout */
		fprintf(fd, "\t\tcylinder { <0,0,%f>, <0,0,-%f>, %f}\n", 
			zoffset+0.01, zoffset+0.01, Gear_Center_Cutout_Radius );
		fprintf(fd, "\t}\n");

		/* The hub and spokes */
		if ( Gear_Hub_Radius != 0.0 )
		{
			fprintf(fd, "\t// Hub and Spokes\n");
			fprintf(fd, "\tcylinder { <0,0,%f>, <0,0,-%f>, %f}\n", 
				zoffset, zoffset, Gear_Hub_Radius );
			rotation_amount = 360.0/(double)Gear_Spoke_Number;
			for ( i=0; i < Gear_Spoke_Number; i++ )
			{
				fprintf(fd, "\tbox { <%f,%f,%f>, <%f,%f,-%f> ", 
					-1*Gear_Hub_Radius/3, 0.0, zoffset, 
					Gear_Hub_Radius/3, Gear_Radius, zoffset);
				fprintf(fd, "\t\trotate z*%f }\n", i*rotation_amount );
			}
		}
	}
	else
		fprintf(fd, "\tcylinder { <0,0,%f>, <0,0,-%f>, %f}\n", 
			zoffset, zoffset, Gear_Radius );

	/*
	 * Now output the teeth.  The arrays are for quadrant 1, so we'll
	 * output the other quadrants using symmetry.
	 *
	 * Each tooth is made up of 4 surfaces:
	 * The front and back surfaces (one triangle in +z, one in -z)
	 * The two side surfaces, each made of two triangles that span +z to -z
 	 *
	 * We'll make the front and back surfaces first, then go back and make
	 * the side surfaces.  This makes for easier reading code here.
	 */

	/* The front and back surfaces */

	fprintf(fd, "\n\t// Front and Back surfaces of teeth\n");

	A0_link = A0;
	A1_link = A1;
	for ( pt=0; pt<Gear_Number_Of_Teeth; pt++ )
	{
		printf(".");

		/* Front face quadrant 1 */
		fprintf( fd, "\ttriangle { <%f,%f,%f>, <%f,%f,%f>, <%f,%f,%f> }\n",
			 A0_link->x, A0_link->y, zoffset,
			 A0_link->next->x, A0_link->next->y, zoffset,
			 A1_link->x, A1_link->y, zoffset );
		/* Back face */
		fprintf( fd, "\ttriangle { <%f,%f,-%f>, <%f,%f,-%f>, <%f,%f,-%f> }\n",
			 A0_link->x, A0_link->y, zoffset,
			 A0_link->next->x, A0_link->next->y, zoffset,
			 A1_link->x, A1_link->y, zoffset );

		/* Front face quadrant 2 */
		fprintf( fd, "\ttriangle { <%f,%f,%f>, <%f,%f,%f>, <%f,%f,%f> }\n",
			 A0_link->x, -1*A0_link->y, zoffset,
			 A0_link->next->x, -1*A0_link->next->y, zoffset,
			 A1_link->x, -1*A1_link->y, zoffset );
		/* Back face */
		fprintf( fd, "\ttriangle { <%f,%f,-%f>, <%f,%f,-%f>, <%f,%f,-%f> }\n",
			 A0_link->x, -1*A0_link->y, zoffset,
			 A0_link->next->x, -1*A0_link->next->y, zoffset,
			 A1_link->x, -1*A1_link->y, zoffset );

		/* Front face quadrant 3 */
		fprintf( fd, "\ttriangle { <%f,%f,%f>, <%f,%f,%f>, <%f,%f,%f> }\n",
			 -1*A0_link->x, -1*A0_link->y, zoffset,
			 -1*A0_link->next->x, -1*A0_link->next->y, zoffset,
			 -1*A1_link->x, -1*A1_link->y, zoffset );
		/* Back face */
		fprintf( fd, "\ttriangle { <%f,%f,-%f>, <%f,%f,-%f>, <%f,%f,-%f> }\n",
			 -1*A0_link->x, -1*A0_link->y, zoffset,
			 -1*A0_link->next->x, -1*A0_link->next->y, zoffset,
			 -1*A1_link->x, -1*A1_link->y, zoffset );

		/* Front face quadrant 4 */
		fprintf( fd, "\ttriangle { <%f,%f,%f>, <%f,%f,%f>, <%f,%f,%f> }\n",
			 -1*A0_link->x, A0_link->y, zoffset,
			 -1*A0_link->next->x, A0_link->next->y, zoffset,
			 -1*A1_link->x, A1_link->y, zoffset );
		/* Back face */
		fprintf( fd, "\ttriangle { <%f,%f,-%f>, <%f,%f,-%f>, <%f,%f,-%f> }\n",
			 -1*A0_link->x, A0_link->y, zoffset,
			 -1*A0_link->next->x, A0_link->next->y, zoffset,
			 -1*A1_link->x, A1_link->y, zoffset );

		A0_link = A0_link->next;
		A1_link = A1_link->next;
	}

	/* The side surfaces */

	fprintf(fd, "\n\t// Side surfaces of teeth\n");

	A0_link = A0;
	A1_link = A1;
	for ( pt=0; pt<Gear_Number_Of_Teeth; pt++ )
	{
		printf(".");

		/* left side, triangle 1 quadrant 1 */
		fprintf( fd, "\ttriangle { <%f,%f,%f>, <%f,%f,%f>, <%f,%f,-%f> }\n",
			 A0_link->x, A0_link->y, zoffset,
			 A1_link->x, A1_link->y, zoffset,
			 A1_link->x, A1_link->y, zoffset );
		/* left side, triangle 2 */
		fprintf( fd, "\ttriangle { <%f,%f,%f>, <%f,%f,-%f>, <%f,%f,-%f> }\n",
			 A0_link->x, A0_link->y, zoffset,
			 A0_link->x, A0_link->y, zoffset,
			 A1_link->x, A1_link->y, zoffset );

		/* left side, triangle 1 quadrant 2 */
		fprintf( fd, "\ttriangle { <%f,%f,%f>, <%f,%f,%f>, <%f,%f,-%f> }\n",
			 A0_link->x, -1*A0_link->y, zoffset,
			 A1_link->x, -1*A1_link->y, zoffset,
			 A1_link->x, -1*A1_link->y, zoffset );
		/* left side, triangle 2 */
		fprintf( fd, "\ttriangle { <%f,%f,%f>, <%f,%f,-%f>, <%f,%f,-%f> }\n",
			 A0_link->x, -1*A0_link->y, zoffset,
			 A0_link->x, -1*A0_link->y, zoffset,
			 A1_link->x, -1*A1_link->y, zoffset );

		/* left side, triangle 1 quadrant 3 */
		fprintf( fd, "\ttriangle { <%f,%f,%f>, <%f,%f,%f>, <%f,%f,-%f> }\n",
			 -1*A0_link->x, -1*A0_link->y, zoffset,
			 -1*A1_link->x, -1*A1_link->y, zoffset,
			 -1*A1_link->x, -1*A1_link->y, zoffset );
		/* left side, triangle 2 */
		fprintf( fd, "\ttriangle { <%f,%f,%f>, <%f,%f,-%f>, <%f,%f,-%f> }\n",
			 -1*A0_link->x, -1*A0_link->y, zoffset,
			 -1*A0_link->x, -1*A0_link->y, zoffset,
			 -1*A1_link->x, -1*A1_link->y, zoffset );

		/* left side, triangle 1 quadrant 4 */
		fprintf( fd, "\ttriangle { <%f,%f,%f>, <%f,%f,%f>, <%f,%f,-%f> }\n",
			 -1*A0_link->x, A0_link->y, zoffset,
			 -1*A1_link->x, A1_link->y, zoffset,
			 -1*A1_link->x, A1_link->y, zoffset );
		/* left side, triangle 2 */
		fprintf( fd, "\ttriangle { <%f,%f,%f>, <%f,%f,-%f>, <%f,%f,-%f> }\n",
			 -1*A0_link->x, A0_link->y, zoffset,
			 -1*A0_link->x, A0_link->y, zoffset,
			 -1*A1_link->x, A1_link->y, zoffset );


		/* right side, triangle 1 quadrant 1 */
		fprintf( fd, "\ttriangle { <%f,%f,%f>, <%f,%f,%f>, <%f,%f,-%f> }\n",
			 A0_link->next->x, A0_link->next->y, zoffset,
			 A1_link->x, A1_link->y, zoffset,
			 A1_link->x, A1_link->y, zoffset );
		/* right side, triangle 2 */
		fprintf( fd, "\ttriangle { <%f,%f,%f>, <%f,%f,-%f>, <%f,%f,-%f> }\n",
			 A0_link->next->x, A0_link->next->y, zoffset,
			 A0_link->next->x, A0_link->next->y, zoffset,
			 A1_link->x, A1_link->y, zoffset );

		/* right side, triangle 1 quadrant 2 */
		fprintf( fd, "\ttriangle { <%f,%f,%f>, <%f,%f,%f>, <%f,%f,-%f> }\n",
			 A0_link->next->x, -1*A0_link->next->y, zoffset,
			 A1_link->x, -1*A1_link->y, zoffset,
			 A1_link->x, -1*A1_link->y, zoffset );
		/* right side, triangle 2 */
		fprintf( fd, "\ttriangle { <%f,%f,%f>, <%f,%f,-%f>, <%f,%f,-%f> }\n",
			 A0_link->next->x, -1*A0_link->next->y, zoffset,
			 A0_link->next->x, -1*A0_link->next->y, zoffset,
			 A1_link->x, -1*A1_link->y, zoffset );

		/* right side, triangle 1 quadrant 3 */
		fprintf( fd, "\ttriangle { <%f,%f,%f>, <%f,%f,%f>, <%f,%f,-%f> }\n",
			 -1*A0_link->next->x, -1*A0_link->next->y, zoffset,
			 -1*A1_link->x, -1*A1_link->y, zoffset,
			 -1*A1_link->x, -1*A1_link->y, zoffset );
		/* right side, triangle 2 */
		fprintf( fd, "\ttriangle { <%f,%f,%f>, <%f,%f,-%f>, <%f,%f,-%f> }\n",
			 -1*A0_link->next->x, -1*A0_link->next->y, zoffset,
			 -1*A0_link->next->x, -1*A0_link->next->y, zoffset,
			 -1*A1_link->x, -1*A1_link->y, zoffset );

		/* right side, triangle 1 quadrant 4 */
		fprintf( fd, "\ttriangle { <%f,%f,%f>, <%f,%f,%f>, <%f,%f,-%f> }\n",
			 -1*A0_link->next->x, A0_link->next->y, zoffset,
			 -1*A1_link->x, A1_link->y, zoffset,
			 -1*A1_link->x, A1_link->y, zoffset );
		/* right side, triangle 2 */
		fprintf( fd, "\ttriangle { <%f,%f,%f>, <%f,%f,-%f>, <%f,%f,-%f> }\n",
			 -1*A0_link->next->x, A0_link->next->y, zoffset,
			 -1*A0_link->next->x, A0_link->next->y, zoffset,
			 -1*A1_link->x, A1_link->y, zoffset );

		A0_link = A0_link->next;
		A1_link = A1_link->next;
	}

	printf("\n");

	/* 
	 * Close up the declaration
	 */
	fprintf(fd, "\t}\n");

	/*
	 * Close the output file
	 */
	fclose(fd);
	return;
}
